-- https://en.wikipedia.org/wiki/Space-filling_curve
-- https://en.wikipedia.org/wiki/Maze_generation_algorithm

-- Disable output buffer so debug messages print in real time
io.stdout:setvbuf("no")

function love.load()
    bit = require "bit"
    math.randomseed(os.time())
    math.random(); math.random(); math.random()
    love.window.setTitle("Space-filling curve Dithering")
    love.graphics.setDefaultFilter("nearest", "nearest", 1)
    imgData = love.image.newImageData("David.png")
    origImg = love.graphics.newImage(imgData)
    scale = 6
    imgSize = imgData:getWidth()
    love.window.setMode(imgSize * scale * 2, imgSize * scale)

    threshold = 0.5 -- The threshold for seperating white and black pixels
    drawGridToggle = false
    animateDither = true
    animationDelay = 20
    timer = 0

    currentCurve = setWallsZero
    currentCurve()
end

function love.keypressed(key)
    if key == "f1" then
        currentCurve = setWallsZero
        currentCurve()
    end
    if key == "f2" then
        currentCurve = setWallsMax
        currentCurve()
    end
    if key == "f3" then
        currentCurve = setWallsRandom
        currentCurve()
    end
    if key == "f4" then
        currentCurve = setWallsBacktracker
        currentCurve()
    end
    if key == "f5" then
        currentCurve = setWallsHilbertFollow
        currentCurve()
    end
    if key == "f6" then
        currentCurve = setWallsHilbert
        currentCurve()
    end
    if key == "t" then
        drawGridToggle = not drawGridToggle
    end
    if key == "a" then
        animateDither = not animateDither
    end
    if key == "escape" then
        love.event.quit()
    end
end

function resetCellGrid()
    -- cellGrid will hold a number of allowed directions, 1 = up, 2 = right, 4 = down, 8 = left
    -- 0 is unvisited
    -- the second entry is the red from our picture, reset this each run as we mess with cellGrid when dithering
    cellGrid = {}
    for x = 0, imgSize - 1 do
        cellGrid[x] = {}
        for y = 0, imgSize - 1 do
            cellGrid[x][y] = {0, imgData:getPixel(x, y)}
        end
    end
end

function setWallsZero()
    resetCellGrid()
    -- set no passability, essentialy threshold shading
    for x = 0, imgSize - 1 do
        for y = 0, imgSize - 1 do
            cellGrid[x][y][1] = 0
        end
    end
    if ditheredImg then ditheredImg:release() end
    ditheredImg = applyDither()
end

function setWallsMax()
    resetCellGrid()
    -- set all passability, allowing error to travel forever
    for x = 0, imgSize - 1 do
        for y = 0, imgSize - 1 do
            cellGrid[x][y][1] = 15
            if x == imgSize - 1 then
                cellGrid[x][y][1] = bit.band(cellGrid[x][y][1], 13)
            end
            if y == imgSize - 1 then
                cellGrid[x][y][1] = bit.band(cellGrid[x][y][1], 11)
            end
        end
    end
    if ditheredImg then ditheredImg:release() end
    ditheredImg = applyDither()
end

function setWallsRandom()
    resetCellGrid()
    -- set passability randomly, block off last row and column
    for x = 0, imgSize - 1 do
        for y = 0, imgSize - 1 do
            cellGrid[x][y][1] = math.random(16)
            if x == imgSize - 1 then
                cellGrid[x][y][1] = bit.band(cellGrid[x][y][1], 13)
            end
            if y == imgSize - 1 then
                cellGrid[x][y][1] = bit.band(cellGrid[x][y][1], 11)
            end
        end
    end
    if ditheredImg then ditheredImg:release() end
    ditheredImg = applyDither()
end

function setWallsBacktracker()
    resetCellGrid()
    -- set passability with a recursive backtracker
    local currX, currY = 0, 0
    local neighbours, r, dir
    local stack = {{currX, currY}}
    while #stack > 0 do
        -- get neightbours
        neighbours = getNeighbours(currX, currY)
        if #neighbours == 0 then
            -- we have no eligible neighbours, step back
            table.remove(stack)
            if #stack > 0 then
                currX, currY = stack[#stack][1], stack[#stack][2]
            end
        else
            -- we have neighbours, get a random one
            r = math.random(#neighbours)
            dir = neighbours[r][3]
            -- set our exit
            cellGrid[currX][currY][1] = bit.bor(cellGrid[currX][currY][1], dir)
            -- set new current
            currX = neighbours[r][1]
            currY = neighbours[r][2]
            table.insert(stack, {currX, currY})
            -- set our entry
            if dir == 1 then
                dir = 4
            elseif dir == 2 then
                dir = 8
            elseif dir == 4 then
                dir = 1
            else
                dir = 2
            end
            cellGrid[currX][currY][1] = bit.bor(cellGrid[currX][currY][1], dir)
        end
    end
    if ditheredImg then ditheredImg:release() end
    ditheredImg = applyDither()
end

function setWallsHilbert()
    resetCellGrid()
    if ditheredImg then ditheredImg:release() end
    local hText = require "hilbert"
    local dir = 2
    local currX, currY = 0, 127
    local char
    for i = 1, #hText do
        char = hText:sub(i, i)
        if char == "-" then
            -- turn left
            dir = dir / 2
            if dir < 1 then dir = 8 end
        elseif char == "+" then
            -- turn right
            dir = dir * 2
            if dir > 8 then dir = 1 end
        end
        --move forwards
        cellGrid[currX][currY][1] = bit.bor(cellGrid[currX][currY][1], dir)
        col = cellGrid[currX][currY][2]
        if dir == 1 then
            currY = currY - 1
            cellGrid[currX][currY][1] = bit.bor(cellGrid[currX][currY][1], 4)
        elseif dir == 2 then
            currX = currX + 1
            cellGrid[currX][currY][1] = bit.bor(cellGrid[currX][currY][1], 8)
        elseif dir == 4 then
            currY = currY + 1
            cellGrid[currX][currY][1] = bit.bor(cellGrid[currX][currY][1], 1)
        else
            currX = currX - 1
            cellGrid[currX][currY][1] = bit.bor(cellGrid[currX][currY][1], 2)
        end
    end
    if ditheredImg then ditheredImg:release() end
    ditheredImg = applyDither()
end

function setWallsHilbertFollow()
    resetCellGrid()
    if ditheredImg then ditheredImg:release() end
    local col, err
    local newImgData = love.image.newImageData(imgSize, imgSize)
    local hText = require "hilbert"
    local dir = 2
    local currX, currY = 0, 127
    local char
    for i = 1, #hText do
        char = hText:sub(i, i)
        if char == "-" then
            -- turn left
            dir = dir / 2
            if dir < 1 then dir = 8 end
        elseif char == "+" then
            -- turn right
            dir = dir * 2
            if dir > 8 then dir = 1 end
        end
        --move forwards
        cellGrid[currX][currY][1] = bit.bor(cellGrid[currX][currY][1], dir)
        col = cellGrid[currX][currY][2]
        -- set cell and get error
        if col < threshold then
            err = col
            col = 0
        else
            err = col - 1
            col = 1
        end
        -- draw pixel
        newImgData:setPixel(currX, currY, col, col, col, 1)
        if dir == 1 then
            currY = currY - 1
            cellGrid[currX][currY][1] = bit.bor(cellGrid[currX][currY][1], 4)
        elseif dir == 2 then
            currX = currX + 1
            cellGrid[currX][currY][1] = bit.bor(cellGrid[currX][currY][1], 8)
        elseif dir == 4 then
            currY = currY + 1
            cellGrid[currX][currY][1] = bit.bor(cellGrid[currX][currY][1], 1)
        else
            currX = currX - 1
            cellGrid[currX][currY][1] = bit.bor(cellGrid[currX][currY][1], 2)
        end
        -- push error
        cellGrid[currX][currY][2] = cellGrid[currX][currY][2] + err
    end
    ditheredImg = love.graphics.newImage(newImgData)
end

function getNeighbours(x, y)
    local neighbours = {}
    -- up
    if y > 0 then
        if cellGrid[x][y - 1][1] == 0 then
            table.insert(neighbours, {x, y - 1, 1})
        end
    end
    -- right
    if x < imgSize - 1 then
        if cellGrid[x + 1][y][1] == 0 then
            table.insert(neighbours, {x + 1, y, 2})
        end
    end
    -- down
    if y < imgSize - 1 then
        if cellGrid[x][y + 1][1] == 0 then
            table.insert(neighbours, {x, y + 1, 4})
        end
    end
    -- left
    if x > 0 then
        if cellGrid[x - 1][y][1] == 0 then
            table.insert(neighbours, {x - 1, y, 8})
        end
    end
    return neighbours
end

function applyDither()
    local col, err, val
    local newImgData = love.image.newImageData(imgSize, imgSize)
    for x = 0, imgSize - 1 do
        for y = 0, imgSize - 1 do
            val = cellGrid[x][y][1]
            col = cellGrid[x][y][2]
            -- set cell and get error
            if col < threshold then
                err = col
                col = 0
            else
                err = col - 1
                col = 1
            end
            -- get neighbours to pass error
            -- both
            if bit.band(val, 6) == 6 then
                err = err / 2
                cellGrid[x + 1][y][2] = cellGrid[x + 1][y][2] + err
                cellGrid[x][y + 1][2] = cellGrid[x][y + 1][2] + err
            -- right
            elseif bit.band(val, 2) == 2 then
                cellGrid[x + 1][y][2] = cellGrid[x + 1][y][2] + err
            -- down
            elseif bit.band(val, 4) == 4 then
                cellGrid[x][y + 1][2] = cellGrid[x][y + 1][2] + err
            end
            -- set the image data
            newImgData:setPixel(x, y, col, col, col, 1)
        end
    end
    return love.graphics.newImage(newImgData)
end

function drawGrid()
    -- draw grid, only need to draw right and down
    local val
    love.graphics.setLineWidth(1 / scale)
    for x = 0, imgSize - 1 do
        for y = 0, imgSize - 1 do
            val = cellGrid[x][y][1]
            -- check if we are blocked right
            if bit.band(val, 2) == 0 then
                love.graphics.line(x + 1, y, x + 1, y + 1)
            end
            -- check if we are blocked down
            if bit.band(val, 4) == 0 then
                love.graphics.line(x, y + 1, x + 1, y + 1)
            end
        end
    end
end

function love.update(dt)
    timer = timer + 1
    if timer % animationDelay == 0 and animateDither then
        currentCurve()
    end
end

function love.draw()
    local start = love.timer.getTime()

    love.graphics.push()
    love.graphics.scale(scale)
    love.graphics.draw(origImg)
    if drawGridToggle == true then
        love.graphics.setColor(0.2, 0.4, 0.8)
        drawGrid()
        love.graphics.setColor(1, 1, 1)
    end
    love.graphics.draw(ditheredImg, imgSize, 0)

    love.graphics.pop()
    -- ui
    love.graphics.print([[F1 - All walls blocked
F2 - All walls open
F3 - Random walls
F4 - Recursive backtracker maze
F5 - Hilbert Curve (following curve)
F6 - Hilbert Curve (from top left)

T - Toggle grid
A - Toggle animation]], 5, 5)
end